/* $Header: /usr/files/home/Gamma/vs/tv/edit/lib/tv/RCS/vsCurFit.c,v 1.3 1994/12/08 20:44:11 vs Exp vs $ */
/* TOOLBOX allowing enhanced fitting of data sets by using the Marquardt-
 * Levenberg - algorithm. Is mainly derived from LEONE.
 *  Sources :
 *     LEONE ( FORTRAN Code )  by Hanewinkel, Clauberg, Wesselborg, Albers
 *       (Institute of Nuclear Physics Cologne). 
 *     Bevington, Philip R. 1969,
 *       'Data Reduction and Error Analysis for the Physical Sciences'
 *	(New York: McGraw-Hill), Chapter 11.
 *     Press, H. W.; Flannery, B. P.; Teukolsky S. A.; Vetterling W. T. 1986
 *       'Numerical Recipes'
 *	(Cambridge: Cambridge University Press)
 *     Marquardt, D.W., 1963, J.Soc.Ind.Appl.Math.,vol.11,pp.431-441
 *     Jacobs, David A.H.,ed.1977,'The State of the Art in Numerical Analysis'
 *       (London: Academic Press),chapter III.2 (by J.E.Dennis).
 *     Weber, Markus, 1987, Kap 10 pg 228ff, (Vieweg Braunschweig)
 ---------------------------------------------------------------------------*/

#define _vsCurFit_c

#include "vsTypes.h"

#include "vsAlloc.h"
#include "vsPar.h"
#include "vsTools.h"

#include "vsCurFit.h"
#include "vsCurFitP.h"

/*---------------------------------------------------------------------------*/
static Cur_t numericCp;
/*---------------------------------------------------------------------------*/
char *CurError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* CURNOERR:  */ "vsCurFit: no error",
  /* CURPOINT:  */ "vsCurFit: illegal NULL-pointer",
  /* CUR0FREE:  */ "vsCurFit: no free parameter",
  /* CUR0DATA:  */ "vsCurFit: no data points to fit",
  /* CURLDATA:  */ "vsCurFit: no degree of freedom",
  /* CURMPAR:   */ "vsCurFit: too much degrees of freedom",
  /* CURFCPNT:  */ "vsCurFit: illegal function NULL-pointer",
  /* CURSING:   */ "vsCurFit: fit matrix has become singular",
  /* CURSING2:  */ "vsCurFit: fit matrix has become singular-2",
  /* CURALLOC:  */ "vsCurFit: can not allocate temporary buffers",
  /* CURINDEX:  */ "vsCurFit: can not get free parameters",
  /* default:   */ "vsCurFit: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/*---------------------------------------------------------------------------*/
int CurFree(cp)
     Cur_t cp;
{
  if (cp == NULL) { CurErrNo = CURPOINT; return FALSE; }
  else {
    ParFree(CurTry(cp));
    VecFree(CurCovar(cp));
    VecFree(CurData(cp));
    return TRUE; }    
}
/*---------------------------------------------------------------------------*/
int CurAlloc(cp, dataSiz)
     Cur_t cp;
     int dataSiz;
{
  if (cp == NULL) { CurErrNo = CURPOINT; return FALSE; }
  else { return VecAlloc(CurData(cp), dataSiz); }
}
/*---------------------------------------------------------------------------*/
int CurFit(cp)
     Cur_t cp;
{
  if (!CurSetup(cp) || !CurCheck(cp)) { return FALSE; }
  else {
    double tryMeasure;
    double oldMeasure;
    double measureRatio;
    double beta[CUR_MAXORDER];
    double oldBeta[CUR_MAXORDER];
    double oldDiag[CUR_MAXORDER];
    Par_t pp= CurPar(cp);
    Par_t tp= CurTry(cp);
    double *alpha= VecP(CurCovar(cp));
    double (*measure)();
    double cmp;

    if (CurMeasureTyp(cp) == CUR_POI) {
      measure= CurPoissonLikelihood;
      cmp= -1.0; }		/* likelihood becomes maximized */
    else {
      measure= CurChiSqr;
      cmp= 1.0; }		/* chi**2 becomes minimized */

    numericCp= cp;
    tryMeasure = CurMeasure(cp) = measure(cp, pp);
    if (!CurEvalCurvature(cp, alpha, oldBeta, oldDiag)) { return FALSE; }

    CurIterations(cp) = 0;

    do {			/* do at least one iteration! */
      int getting_better= FALSE;

      if (CurProtocol(cp) != NULL) { (*CurProtocol(cp))(cp); }

      oldMeasure = CurMeasure(cp);
      while (!getting_better) {
	{ int i;
	  for (i= 0; i < CurFreeParNum(cp); i++) {
	    double lambda= CurLambda(cp);
	    beta[i] = oldBeta[i];
	    alpha[ i*(i+3)/2 ] = oldDiag[i] * (1. + lambda); } }
	if (!CurSolveSymmetric(alpha, beta, CurFreeParNum(cp))){
	  return FALSE; }
	(*CurChangeTry(cp))(cp, beta);
	tryMeasure = measure(cp, tp);
	
	if ((CurMeasureTyp(cp) != CUR_POI && tryMeasure == 0.0) ||
	    cmp * tryMeasure < cmp * oldMeasure) { /* getting closer... */
	  /*CurLambda(cp) *= CurDecLambda(cp);*/
				/*that's a better way: */
	  CurLambda(cp) = .001 * (oldMeasure - tryMeasure);
	  getting_better = TRUE; }
	else {
	  CurLambda(cp) *= CurIncLambda(cp);
	  if (CurLambda(cp) > CurMaxLambda(cp)) {
				/* two cases: i. minimum is already reached */
				/*           ii. skipped minimum */
				/* if ii. try a half step back to get better */
	    { int i;
	      for (i= 0; i< CurFreeParNum(cp); i++) { beta[i] *= (-.5); } }
	    (*CurChangeTry(cp))(cp, beta);
	    tryMeasure = measure(cp, tp);
	    if (cmp * tryMeasure < cmp * oldMeasure) { /* Aha case ii. !!! */
	      getting_better = TRUE; }
	    else {	    	/* get out of here! */
	      getting_better = FALSE; /* ==> measureRatio = 0. */
	      break; } } } }
      if (getting_better) {	/* new parameter set, beta and measure */
	ParCopy(pp, tp);
	if (!CurEvalCurvature(cp, alpha, oldBeta, oldDiag)) { return FALSE; }
	CurMeasure(cp) = tryMeasure; }
      if (CurLambda(cp) < CurMinLambda(cp)) {
	CurLambda(cp) = CurMinLambda(cp);}
      measureRatio = ((CurMeasure(cp) == 0.) ? 0. :
		      (cmp * (oldMeasure - CurMeasure(cp)) / CurMeasure(cp)));

      CurIterations(cp)++;
    } while ((CurIterations(cp)< CurMaxIterations(cp) &&
	      measureRatio > CurMeasureRatio(cp)) ||
	     (CurMeasureTyp(cp) == CUR_POI && CurIterations(cp) == 1)) ;

    return CurFinish(cp, oldBeta, beta, oldDiag); }
}
/*---------------------------------------------------------------------------*/
int CurInit(cp)
     Cur_t cp;
{
  if (cp == NULL) { CurErrNo = CURPOINT; return FALSE; }
  else {
    ParInit(CurTry(cp));
    VecInit(CurCovar(cp), sizeof(double));
    VecInit(CurData(cp), sizeof(CurPnt_t));
    CurReset(cp);
    return TRUE; }
}
/*---------------------------------------------------------------------------*/
int CurReset(cp)
     Cur_t cp;
{
  if (cp == NULL) { CurErrNo = CURPOINT; return FALSE; }
  else {
    ParReset(CurTry(cp));
    VecReset(CurCovar(cp));
    VecReset(CurData(cp));

    CurFreedom(cp)		= 0;
    CurMeasure(cp)		= CurDMeasure(cp) = 0.0;
    CurMeasureRatio(cp)		= CUR_RATIO;
    CurLambda(cp)		= CUR_LAMBDA;
    CurMinLambda(cp)		= CUR_MINLAMBDA;
    CurMaxLambda(cp)		= CUR_MAXLAMBDA;
    CurDecLambda(cp)		= CUR_DECLAMBDA;
    CurIncLambda(cp)		= CUR_INCLAMBDA;
    CurIterations(cp)		= 0;
    CurMaxIterations(cp)	= CUR_ITERATIONS;
    CurFunctionSetup(cp)	= NULL;
    CurFunctionFinish(cp)	= NULL;
    CurFunction(cp)		= NULL;
    CurDerivationSetup(cp)	= CurNumericDerivationOk;
    CurDerivationFinish(cp)	= CurNumericDerivationOk;
    CurDerivation(cp)		= CurNumericDerivation;
    CurProtocol(cp)		= NULL;
    CurChangeTry(cp)		= NULL;
    return TRUE; }
}
/*---------------------------------------------------------------------------*/
int CurSetup(cp)
     Cur_t cp;
{
  if (cp == NULL) { CurErrNo = CURPOINT; return FALSE; }
  else if (!ParIndex(CurPar(cp))) { CurErrNo= CURINDEX; return FALSE; }
  else {
    Par_t pp= CurPar(cp);
    int size = ParNumFree(pp) * (ParNumFree(pp) + 3) / 2;
    if (VecAlloc(CurCovar(cp), size) &&
	ParCopy(CurTry(cp), pp)) {
      VecNum(CurCovar(cp)) = size;
      CurFreedom(cp)= CurDataNum(cp) - CurFreeParNum(cp);
      if (CurMeasureRatio(cp)	<= 0.0) { CurMeasureRatio(cp)= CUR_RATIO; }
      if (CurLambda(cp)		<= 0.0) { CurLambda(cp)    = CUR_LAMBDA; }
      if (CurMinLambda(cp)	<= 0.0) { CurMinLambda(cp) = CUR_MINLAMBDA; }
      if (CurMaxLambda(cp)	<= 0.0) { CurMaxLambda(cp) = CUR_MAXLAMBDA; }
      if (CurDecLambda(cp)	<= 0.0) { CurDecLambda(cp) = CUR_DECLAMBDA; }
      if (CurIncLambda(cp)	<= 0.0) { CurIncLambda(cp) = CUR_INCLAMBDA; }
      return TRUE; }
    else { CurErrNo= CURALLOC; return FALSE; } }
}
/*---------------------------------------------------------------------------*/
static int CurCheck(cp)
     Cur_t cp;
{
  if (CurFreeParNum(cp) <= 0) { CurErrNo= CUR0FREE; }
  else if (CurDataNum(cp) <= 0) { CurErrNo= CUR0DATA; }
  else if (CurFreedom(cp) <= 0) { CurErrNo= CURLDATA; }
  else if (CurFreeParNum(cp) > CUR_MAXORDER) { CurErrNo= CURMPAR; }
  else if (CurFunctionSetup(cp) == NULL || CurFunctionFinish(cp) == NULL ||
	   CurDerivationSetup(cp) == NULL || CurDerivationFinish(cp) == NULL ||
	   CurFunction(cp) == NULL ||
	   CurDerivation(cp) == NULL ||
	   CurChangeTry(cp) == NULL) { CurErrNo= CURFCPNT; }
  else { return TRUE; }

  return FALSE;
}
/*---------------------------------------------------------------------------*/
static double CurChiSqr(cp, pp)
     Cur_t cp;
     Par_t pp;
{
  register double chi= 0.0;

  if (CurFunctionSetup(cp)(pp)) {
    register int i;
    register CurPntVec_t cdp= CurData(cp);

    for (i= 0; i< VecNum(cdp); i++) {
      register double f = (*CurFunction(cp))(CurPntVecX(cdp,i));
      register double c = CurPntVecY(cdp,i) - f;
      chi += CurPntVecW(cdp,i) * c * c; }

    chi /= CurFreedom(cp);
    CurFunctionFinish(cp)(); }
  return chi;
}
/*---------------------------------------------------------------------------*/
static double CurPoissonLikelihood(cp, pp)
     Cur_t cp;
     Par_t pp;
{
  register double like= 0.0;

  if (CurFunctionSetup(cp)(pp)) {
    register int i;
    register CurPntVec_t cdp= CurData(cp);

    for (i= 0; i< VecNum(cdp); i++) {
      register double f = (*CurFunction(cp))(CurPntVecX(cdp,i));
      register double lf= (f > 0.0) ? log(f) : -1000.0;
      like += CurPntVecY(cdp,i) * lf - f; }
    CurFunctionFinish(cp)(); }
  return like;
}
/*---------------------------------------------------------------------------*/
static int CurEvalCurvature(cp, alpha, oldBeta, oldDiag)
				/* evaluate curvature matrix and gradient */
				/* 21.07.89 (sk) */
     /* explain : if you look below it may look a bit tricky, so I need to
        help you. first of all you should know that 'a' is a symmetric matrix
	with Free_par_no * (Free_par_no + 1) / 2 different (!) elements.
	why should I calculate elements that I already know? I don't.
	'a' corresponds as follows to the matrix:
	   a[0]  a[1]  a[3]  a[6] ...    
	   a[1]  a[2]  a[4]  a[7] ...    
	   a[3]  a[4]  a[5]  ...
	   a[6]  a[7]  ...
	   ...
	as you can see easily the diagonal elements are found at
	   j * (j + 3) / 2     :  j = 0 ... Free_par_no-1
	*/
     Cur_t cp;
     double *alpha;
     double *oldBeta;
     double *oldDiag;
{
  memset((void *)alpha, 0,
	 (CurFreeParNum(cp) + 3) * CurFreeParNum(cp) / 2 *sizeof(double));
  memset((void *)oldBeta, 0, CurFreeParNum(cp) * sizeof(double));

  if (CurDerivationSetup(cp)(CurPar(cp))) {
    CurPntVec_t cdp= CurData(cp);
    { int i;
      for (i= 0; i< VecNum(cdp); i++) {
	double grad_a[CUR_MAXORDER];
	double sigm;
	double fx;
	double dy;		/* y - f(x) */
	CurDerivation(cp)(CurPntVecX(cdp,i), grad_a);
	fx = CurFunction(cp)(CurPntVecX(cdp,i));
	dy = CurPntVecY(cdp,i) - fx;

	switch (CurMeasureTyp(cp)) {	/* measure */
	case CUR_POI:
	  sigm= ((CurIterations(cp) ? ABS(fx) : ABS(CurPntVecY(cdp,i))) +
		 CurPntVecO(cdp,i));
	  sigm= (sigm == 0.0) ? 1.0 : (1.0 / sigm);
	  break;
	default:
	  sigm = CurPntVecW(cdp,i);
	  break; }

	{ int j;
	  double *ap, *bp;
	  for (j=0,ap=alpha,bp=oldBeta;j<CurFreeParNum(cp);j++) {
	    int k;
	    double *deri;
	    double wei = grad_a[j] * sigm;
	    *(bp++) += dy * wei;
	    for (k=0,deri=grad_a;k<=j;k++) {
	      *(ap++) += *(deri++) * wei; } } } } }

    { int i;
      for (i=0;i<CurFreeParNum(cp);i++) { /* save diagonal elements */
	*(oldDiag++) = alpha[ i*(i+3)/2 ]; }
    CurDerivationFinish(cp)(); }
    return TRUE; }
  else { return FALSE; }
}
/*---------------------------------------------------------------------------*/
static int CurSolveSymmetric(a, b, order)
     double *a;			/* symmetric matrix  */
     double *b;			/* inhomogenity */
     int  order;		/* order of symmetric matrix system */
{
  register int i, j;
  double matr[CUR_MAXORDER*CUR_MAXORDER];
  for (i=0;i<order;i++) {
    for (j=0;j<=i;j++,a++) {
      *(matr + (i * order) + j) = *a;
      *(matr + (j * order) + i) = *a; } }
  return CurGaussJordan(matr, order, order, b, 1, 1);
}
/*---------------------------------------------------------------------------*/
static int CurInverseSymmetric(a, b, order)
     double *a;
     double *b;
     int  order;
{
  register int i, j;
  double *m=a;
  double matr[CUR_MAXORDER*CUR_MAXORDER];
  for (i=0;i<order;i++) {
    for (j=0;j<=i;j++,m++) {
      *(matr + (i * order) + j) = *m;
      *(matr + (j * order) + i) = *m; } }
  if (CurGaussJordan(matr, order, order, b, 1, 1)) {
    for (i=0;i<order;i++) {
      for (j=0;j<=i;j++,a++) {
	*a = *(matr + (i * order) + j); } }
    return TRUE; }
  else { return FALSE; }
}
/*---------------------------------------------------------------------------*/
int CurGaussJordan(a, n, np, b, m, mp)
				/* solves a lin. eq. syst. by gauss-jordan */
				/* with full pivoting */
				/* 18.07.89 (sk) */
     int n;			/* order of system to solve */
     int np;			/* distance between two rows in units double */
     int m;			/* number of b vectors to solve */
     int mp;			/* physical distance between two col's */
     double *a;			/* given coeff' matrix to solve */
     double *b;			/* matrix of inhomogenity vectors  */
{
  int i, j, k, l, row=0, col=0;
  double big;
  int pivot[CUR_MAXORDER];
  int indexCol[CUR_MAXORDER];
  int indexRow[CUR_MAXORDER];
#define SWAP(x,y) { double tmp= (x); (x)= (y); (y)= tmp; }
#define A(r,c)    (*(a + (r) * np + (c)))	/* use pointer access */
#define B(r,c)    (*(b + (r) * mp + (c)))

  for (i= 0; i< n; i++) { pivot[i] = 0; }
  for (i= 0; i< n; i++) {
    big = 0.;
    for (j= 0; j< n; j++) {
      if (pivot[j] != 1) {
	for (k= 0; k< n; k++){
	  if (pivot[k] == 0) {
	    if (fabs(A(j,k)) >= big ) {
	      big = fabs(A(j,k));
	      row = j;
	      col = k; }}
	  else if (pivot[k] > 1) {
	    CurErrNo= CURSING; return FALSE; } } } }
    pivot[col]++;

    if (row != col) {
      for (l= 0; l< n; l++){ SWAP(A(row,l), A(col,l)); }
      for (l= 0; l< m; l++){ SWAP(B(row,l), B(col,l)); } }
    indexRow[i] = row;
    indexCol[i] = col;
    if (A(col,col) == 0.){
      CurErrNo= CURSING2; return FALSE; }
    else {
      double pivotInv = 1./A(col,col);
      A(col,col) = 1.;
      for (l= 0; l< n; l++) { A(col,l) *= pivotInv; }
      for (l= 0; l< m; l++) { B(col,l) *= pivotInv; } }
    for (k= 0; k< n; k++){
      if (k != col) {
	double tmp = A(k,col);
	A(k,col) = 0;
	for (l= 0; l< n; l++) { A(k,l) -= A(col,l) * tmp; }
	for (l= 0; l< m; l++) { B(k,l) -= B(col,l) * tmp; } } } }

  for (l= n-1; l>= 0; l--) {
    if (indexRow[l] != indexCol[l]) {
      for (k= 0; k< n; k++) { SWAP(A(k,indexRow[l]), A(k,indexCol[l])); } } }

  return TRUE;
#undef SWAP
#undef A
#undef B
}
/*---------------------------------------------------------------------------*/
static int CurFinish(cp, oldBeta, beta, oldDiag)
     Cur_t cp;
     double *oldBeta;
     double *beta;
     double *oldDiag;	   /* diagonal elements of last successful iteration */
{
  register int i, fpn= CurFreeParNum(cp);

  register Par_t pp= CurPar(cp);

  register double *alpha= VecP(CurCovar(cp)), dmeasure= 0.0;

  for (i= 0; i< CurFreeParNum(cp); i++) {
    beta[i]= oldBeta[i];
    alpha[i*(i+3)/2] = oldDiag[i]; }

  if (!CurInverseSymmetric(alpha, beta, CurFreeParNum(cp))){
    return FALSE; }		/* get covariance matrix */

  for (i=0; i< fpn; i++) {
    double pe = ParErr(pp,ParIfree(pp,i)) = sqrt(fabs(alpha[i*(i+3)/2]));
    dmeasure += oldBeta[i] * oldBeta[i] * pe; }

  CurDMeasure(cp) = dmeasure * 4.0;

  return (ParUpdate(pp) && CurErrNo == CURNOERR);
}
/* ---------------------------------------------------------------------- */
static int CurNumericDerivation(x, grad)
     double x;
     double *grad;
{
  double dx[CUR_MAXORDER];
  Par_t pp= CurPar(numericCp);
  double fx0;
  int i;

  for (i= 0; i< CurFreeParNum(numericCp); i++) { dx[i]= 0.0; }
  fx0= CurFunction(numericCp)(x);
  CurFunctionFinish(numericCp)();
  for (i= 0; i< CurFreeParNum(numericCp); i++) {
    if ((dx[i] = ParVal(pp,ParIfree(pp,i)) * CUR_RELATIVDX) == 0.0) {
      dx[i] = CUR_ABSOLUTDX; }
    if (CurChangeTry(numericCp)(numericCp,dx)) {
      if (CurFunctionSetup(numericCp)(CurTry(numericCp))) {
	grad[i]= ((*CurFunction(numericCp))(x) - fx0) / dx[i];
	dx[i] = -dx[i];
	(*CurChangeTry(numericCp))(numericCp,dx);
	dx[i] = 0.0;
	CurFunctionFinish(numericCp)(); }
      else { return FALSE; } }
    else { return FALSE; } }
  if (!CurFunctionSetup(numericCp)(CurPar(numericCp))) { return FALSE; }
  return TRUE;
}
/* ---------------------------------------------------------------------- */
static int CurNumericDerivationOk() { return TRUE; }
/* ---------------------------------------------------------------------- */
