/* 
 * vsSearch.c
 * peaksearch of sassi library
 *
 * Martin Luig
 *
 * started 17. December 1991
 *
 *
 *
 * copyright: sassi library version 1.0 (c) (1991,1992), M. Luig
 *            Institut of Nuclear Physics, University of Cologne.
 *            The library and the header files may be used free of charge,
 *            provided that the following conditions are met:
 *            1. There is no warranty of any kind for all parts of the library.
 *            2. If the use of the library is connected to any form of
 *               publication, the authors must be cited correctly.
 *            3. The authors prohibit, that other person give away or sell copies
 *               of this software. Copies are given only by the authors under the
 *               following postal address:  Universitaet zu Koeln,
 *               Zuelpicherstr. 77, D-5000 Koeln 41, Fed. Rep. Germany.
 */

#include <stdio.h>
#include <math.h>

#include <tv/vsAlloc.h>

#define _vsSearch_c
#include "vsSearch.h"

#define SQR(_a) ((_a)*(_a))
#define nInt(_x) ((_x) < 0.0 ? (int)((_x)-0.5) : (int)((_x)+0.5))

#define SIGMA_TO_FWHM	 2.35482004503094930000
				/* == 2* SQRT(2* M_LN2) */
#define FWHM_TO_SIGMA	 0.42466090014400953000
				/* == 1./SIGMA_TO_FWHM  */
#define EPSILON  1e-10
#define INFINITE  1e38
#ifndef PI
# define PI  3.141592654
#endif

#define LN_HALB_MAL_4 -2.77258872223978    /*       4 * ln(1/2):   */
#define MAX_WIDTH 50.0

static double Last_Peak;

/*
 *  static double gauss_integral (Z)
 *  computes integral of sqrt(2/pi) * exp (-Z**2/2) from -X to +X
 *  developed from Program 3-5 in
 *  P.R.Bevington: Data Reduction and Error Analysis for the Physical 
 *               Sciences, McGraw-Hill New York 1969
 */
static double gauss_integral (Z)
     double Z;

#define GAUSS_PRECISION 0.0000001
#define F1 0.7071067812             /*      sqrt(0.5): */
#define F2 1.128379167              /*      2/sqrt(pi): */

{
  if (Z > 6.0) {
    return (1.0);
  }
  else if (Z > 0.0) {
    double Y2          = Z * Z;
    double Sum         = Z;
    {
      double Term        = Z;
      double Denominator = 1.0;
      while (Term/Sum > GAUSS_PRECISION) {
	Term /= (Denominator += 2.0);
	Term *= Y2;
	Sum  += Term;
      }
    }
    return (F1 * F2 * (exp (-0.5*Y2)) * Sum);
  }
  else {
    return (0.0);
  }
}

#undef GAUSS_PRECISION
#undef F1            
#undef F2

#define MIN_PROBABILITY 0.01
#define DEFAULT_PROBABILITY 0.9999

/*
 *  static SeaPeak _insert_peak (peak_list, P_spec, BG_spec, C_spec, Start_channel, End_channel,
 *                           min_proB, weight_bg, cal)
 *  called by peak_search and insert_peak; 
 *  tries to find a peak between Start_channel and End_channel 
 *  
 *  returns SeaPeak on success, otherwise NULL;
 *
 *  fits  cut(i)= BG + Amp * exp (c*i**2)   where c= ln(1/2) / width**2
 *
 *  at any point between start_channel and end_channel.
 *  the best fit is defined as that one with the highest probability
 *  (probability integral from -Amp/dAmp to +Amp/dAmp) resp.
 *  if that is equal for several fits (i.e. equal to one), between
 *  those fits that one with the least chi square.
 *
 *  "any point" means several runs with decreasing Step_width until
 *  Step_width<= 'precision'.
 * 
 * detailed description of the algorythm see Diplomarbeit (see top)
 */
static SeaPeak _insert_peak (peak_list, P_spec, BG_spec, C_spec, res,
			     Start_channel, End_channel,
			     min_prob, weight_bg, w_cal, cal, peak_func)
SeaPeakList  peak_list;
double       *P_spec,           /* Peak Spectrum                      */
             *BG_spec,          /* Background Spectrum                */
             *C_spec;           /* Cut Spectrum                       */
int          res;               /* spectrum - resolution              */
int          Start_channel;     /* lowest tried peak position:        */
int          End_channel;       /* highest tried peak position:       */
double       min_prob;
double       weight_bg;
double       (*w_cal)();
Cal_t        cal;
int          (*peak_func)();

{
#define STEP_PRECISION 0.03
  
  /* variables in alphabetic orders
   *  general convention:
   *  suffix '0' indicates the actual fit value,
   *  suffix '1' indicates the value of the best fit
   *  the suffixed variables correspond to the global arrays of the peak 
   *  buffer
   */

  double    Amp0;             /*  Amplitude of fitted peak:     */
  double    Amp1 = 0.0;       /*                                */
  double    Vol1 = 0.0;       /*  Volume of fitted peak:        */
  double    BG0;              /*  background of fitted peak:    */
  double    BG1 = 0.0;        /*                                */
  double    dBG1 = 0.0;       /*                                */
  double    Channel;          /*  mean of fitted gauss peak:    */
  double    Chi1;             /*                                */
  double    D0;               /*  determinants in (6-17):       */
  double    D1;               /*                                */
  double    D2;               /*                                */
  double    dAmp1 = 0.0;      /*                                */
  double    dVol1 = 0.0;      /*  standard deviation of 'Volume' from fit: */
  double    First_channel;    /*  start peak position for fit iteration: */
  SeaPeak   found = NULL;     /*  to be returned on success              */
  double    g[201];           /*  computed gauss peak:                   */
  double    Last_channel;     /*  end peak position for fit iteration:   */
  int       Left_channel;     /*  left end of fit region:                */
  double    N;                /*  number of fit points (region width):   */
  double    Peak_channel=0.0; /*  position of best fitted peak:          */
  double    Prob0;            /*  probability of fit:                    */
  double    Prob1;            /*                                         */
  int       Right_channel;    /*  right end of fit region:               */
  double    Step_width;       /*  step width of position of tried gausspeaks:*/
  double    Width0;           /*  half width of peak:                    */
  double    width;            /*  half width of peak:                    */
                           /* yi=cut(j)=counts in cut corresponding to g(i): */
  

  Chi1          = INFINITE;
  Prob1         = 0.0;
  First_channel = Start_channel;
  Last_channel  = End_channel;
  Step_width    = (Last_channel-First_channel+1)*0.1;

  while (Step_width >= STEP_PRECISION) {
    for (Channel= First_channel; Channel <= Last_channel; Channel +=Step_width) {
      Width0 = (*w_cal)(cal, Channel);
      if (Width0 > MAX_WIDTH) { Width0 = MAX_WIDTH; }


      Left_channel = nInt (Channel - 2.0 * Width0);
      Right_channel= nInt (Channel + 2.0 * Width0);
      if (Left_channel >= 0 &&  Right_channel < res && Width0 > 0) {
	double    Sg  = 0.0;                 /*      sum of g(i):                               */
	double    Sgg = 0.0;                 /*      Sum of g(i) squared:                       */
	double    Sgy = 0.0;                 /*      sum of g(i)*y(i):                          */
	double    Sy  = 0.0;                 /*      sum of y(i):                               */
	double    C   = LN_HALB_MAL_4 / (Width0 * Width0);

	{
	  double q     = (double)Left_channel  - Channel;
	  double q_end = (double)Right_channel - Channel;
	  double *pg  = g;
	  double *pyi = C_spec + Left_channel;

	  for ( ; q <= q_end; q++, pg++, pyi++) {
	    register double temp = exp(C * SQR(q));

	    Sg  += temp;
	    Sgg += (temp * temp);
	    Sy  += *pyi;
	    Sgy += (temp * *pyi);
	    *pg = temp;
	  }
	}

	N = Right_channel - Left_channel + 1;  	/*              number of data points: */
	D0     = N  * Sgg - Sg * Sg;
	D1     = Sgg * Sy  - Sg * Sgy;
	D2     = N  * Sgy - Sg * Sy;
	Amp0   = D2/D0;
	BG0    = D1/D0;
	{
	  double Sg_D0  = Sg  / D0;
	  double Sgg_D0 = Sgg / D0;
 	  double N_D0   = N   / D0;
	  double dAmp0  = 0.0;
	  double dBG0   = 0.0;  
	  double Chi0   = 0.0;

	  {
	    double *pg     = g;
	    double *pP     = P_spec + Left_channel;
	    double *pC     = C_spec + Left_channel;
	    double *pP_end = P_spec + Right_channel + 1;
	    
	    for ( ; pP != pP_end; pg++, pP++, pC++) {
#ifdef BG_EXISTS	  
	      {
		int J = pP - P_spec;
		double dyi    = *pP + weight_bg * BG_spec[J];
	      }
#else
	      double dyi    = *pP;
#endif
	      
	      if (dyi < EPSILON) {
		dyi = EPSILON;
	      }
	      {
		double q  = N_D0 * *pg - Sg_D0;
		dAmp0 += SQR(q)  * dyi;
	      }
	      {
		double q = Sgg_D0 - *pg * Sg_D0;
		dBG0  += SQR(q) * dyi;
	      }
	      {
		double q = *pC - Amp0 * *pg - BG0;
		Chi0  += SQR(q) / dyi;
	      }
	    }
	  }
	  Chi0 /= (N-3); 	      /* reduced chi square */
	  if (dAmp0 <= 0) {
	    dAmp0 = 0.0;
	    dBG0  = 0.0;
	    Prob0 = 0.0;
	  }
	  else {
	    dAmp0 = sqrt (dAmp0);
	    dBG0  = sqrt (dBG0);
	    Prob0 = gauss_integral (Amp0/dAmp0);
	  }
	  if (Prob0 > Prob1 || (Prob0 == Prob1 && Chi0 < Chi1)) {
	    Amp1  = Amp0;
	    dAmp1 = dAmp0;
	    BG1   = BG0;
	    dBG1  = dBG0;
	    Chi1  = Chi0;
	    Prob1 = Prob0;
	    width =  Width0; 
	    Peak_channel= Channel;
	    Vol1 = sqrt (-PI/C)*Amp1;
	    if (Amp1 > 0.0) {
	      dVol1= (dAmp1 / Amp1) * Vol1;
	    }
	  }
	}
      }
    }
    First_channel = Peak_channel-Step_width;
    Last_channel  =  Peak_channel+Step_width;
    Step_width /= 10;
  }
  
  {
    int Insert = ((Prob1 >= min_prob) && (Amp1 > 0.0));
    
    if (Insert) {
      Prob1 *= 100.0;
      if ((found = SeaEnterPeak(peak_list, Peak_channel, Vol1, dVol1, Prob1,
				Amp1, dAmp1, BG1, dBG1,
				Chi1, width * FWHM_TO_SIGMA)) == NULL) {
	/* error string received from enter_peak */
	return (NULL);
      }
      if (peak_func != NULL) {
	if ((*peak_func)(SeaListLast(peak_list))) {
	  SeasetErrNo(SEAUSER);
	  return(NULL);
	}
      }
    }
  }
  SeasetErrNo(SEANOERR);
  return(found);
}
   
#undef STEP_PRECISION
#undef MIN_PROBABILITY



/*
 * SeaPeakList _peak_search (P_spec, BG_spec, C_spec, 
 *                          PS_left, PS_right, weight_bg, 
 *                          def_prob, w_cal, cal, peak_func)
 *
 * fits  cut(i)= BG + Amp * exp (c*i**2)   where c= ln(1/2) / width**2
 *
 * at any channel between start_channel and end_channel.
 * If the probability is greater then 'default_probability',
 * (probability integral from -Amp/dAmp to +Amp/dAmp) then
 * 
 * for some basic ideas see H.W.Schuh, Doktorarbeit Koeln 1980,
 * Program PeakFn
 *
 * Detailed commentation see subroutine Insert_peak from which this
 * routine here has been developed (also for description of variables).
 *
 */

static SeaPeakList _peak_search (P_spec, BG_spec, C_spec, res,
				 PS_left, PS_right,
				 weight_bg, def_prob, w_cal, cal, peak_func)
     double       *P_spec;              /* Peak Spectrum */
     double       *BG_spec;             /* Background Spectrum */
     double       *C_spec;              /* Cut Spectrum */
     int          res;                  /* resolution */
     int          PS_left;	        /* region borders for peak search */
     int          PS_right;	        /* ... */
     double       weight_bg;            /* Background weight */
     double       def_prob;             /*                              */
     double       (*w_cal)();           /* function which returns Width */
     Cal_t        cal;                  /* client data for w_cal */
     int          (*peak_func)();       /* User defined function to manipulate peaks "online" */
{
  SeaPeakList peak_list = NULL;
  int        Left_channel, Right_channel;
  
  double     Amp0, BG0, D0, D1, D2, Chi1;
  int        Channel;
  double     N_D0, Prob0, Prob1, Sg_D0;
  double     dAmp0, Width0;
  
  int Peak_there, improving = TRUE;

  if (peak_list == NULL) {
    if (!(peak_list= SeaCreatePeakList())) { return NULL; } }
  
  if (PS_left == 0 && PS_right == 0) {
    PS_left= 1;
    PS_right= res;
  }
  
  Chi1 = INFINITE;
  Prob1 = 0.0;
  Peak_there = FALSE;
    
  for (Channel = PS_left; Channel <=  PS_right; Channel++) {
    Width0= (*w_cal)(cal, (double)Channel);
    if (Width0 > MAX_WIDTH) { Width0 = MAX_WIDTH; }

    Left_channel = nInt (Channel - 2 * Width0);
    Right_channel = nInt (Channel + 2 * Width0);

    if (Left_channel >= 1 && Right_channel <= res && Width0 > 0)  {
      double g[(int)(MAX_WIDTH) * 4 + 1];
      double Chi0;
      double C   = LN_HALB_MAL_4 / (Width0 * Width0);
      double Sg  = 0.0;
      double Sgg = 0.0;
      double Sy  = 0.0;

      double Sgy = 0.0;
      int    N   = Right_channel - Left_channel + 1;
      {
	double q     = (double)Left_channel  - Channel;
	double q_end = (double)Right_channel - Channel;
	double *pg = g;
	double *pyi = C_spec + Left_channel;

	for ( ; q <= q_end; q++, pg++, pyi++) {
	  register double temp = exp(C * SQR(q));

	  Sg  += temp;
	  Sgg += (temp * temp);
	  Sy  += *pyi;
	  Sgy += (temp * *pyi);
	  *pg = temp;
	}
      }
      D0    = N * Sgg - Sg * Sg;
      D1    = Sgg * Sy  - Sg * Sgy;
      D2    = N * Sgy - Sg * Sy;
      Amp0  = D2/D0;
      BG0   = D1/D0;
      Sg_D0 = Sg/D0;
      N_D0  = N /D0;
      dAmp0 = 0.0;
      Chi0  = 0.0;
      {
	double *pg     = g;
	double *pP     = P_spec + Left_channel;
	double *pC     = C_spec + Left_channel;
	double *pP_end = P_spec + Right_channel + 1;

	for ( ; pP != pP_end; pg++, pP++, pC++) {
#ifdef BG_EXISTS	  
	  {
	    int J = pP - P_spec;
	    double dyi    = *pP + weight_bg * BG_spec[J];
	  }
#else
          double dyi    = *pP;
#endif
	  if (dyi < EPSILON) {
	    dyi = EPSILON;
	  }
	  {
	    double q      = N_D0 * *pg - Sg_D0;
	    dAmp0 += SQR(q) * dyi;
	  }
	  {
	    double q = *pC - Amp0 * *pg - BG0;
	    Chi0  += SQR(q) / dyi;
	  }
	}
      }
      Chi0 /= (N-2);
      if (dAmp0 <= 0) {
	dAmp0 = 0;
	Prob0 = 0.0;
      }
      else {
	dAmp0 = sqrt (dAmp0);
	Prob0 = gauss_integral (Amp0/dAmp0);
      }
      if (Peak_there) {
	if (improving) {
	  improving = (Prob0 > Prob1 || (Prob0 == Prob1 && Chi0 < Chi1));
	  if (!improving) {
	    if ((int)Last_Peak < Channel-Width0-2) {
	      int II = Channel;
	      if (_insert_peak (peak_list, P_spec, BG_spec, C_spec, res,
				II-2, II, weight_bg, def_prob,
				w_cal, cal, peak_func) == NULL) {
		if (SeaErrNo != SEANOPEAK) {
		  SeaFreePeakList(peak_list);
		  return (NULL);
		}
	      }
	    }
	    Peak_there= FALSE;
	  }
	}
      }
      else {
	Peak_there = Prob0 >= def_prob;
	improving  = Peak_there;
      }
      Prob1 = Prob0;
      Chi1  = Chi0;
    }      
  }
  return (peak_list);
}

  
#undef EPSILON  
#undef INFINITE
#undef PI  
#undef LN_HALB_MAL_4 
#undef SQR
#undef nInt
#undef MAX_WIDTH
#undef DEFAULT_PROBABILITY

/*
 *   static double *empty_spectrum(res, left, right)
 *   allocate double array with size res and fill it from left to right with 0.0
 *   returns NULL on error
 */
static double *empty_spectrum(res, left, right)
int res, left, right;
{
  double *empty;

  if (!(empty= (double *)Malloc (res * sizeof(double)))) {
    SeasetErrNo(SEAALLOC); return (NULL); }
  {
    double *end = empty + right;
    double *t = empty + left;
    
    for ( ; t <= end ; t++) { *t = 0.0; }
  }
  return (empty);
}
  
#define _CONVERT {                                                    \
                   double *temp = new;                                \
                   double *new_end = new + res;                       \
                   _spec *p = (_spec *)spec;                          \
                   for ( ; temp != new_end; temp++, p++)  {           \
                     *temp = (double) *p;                             \
                   }                                                  \
                 }                                                    \
/* 
 *  static double *convert(spec, res, spec_type)
 *  converts a spectrum of type spec_type to double
 *  returns NULL on error
 */
static double *convert(spec, res, spec_type)
     void *spec;
     int  res;
     int  spec_type;
{
  double *new;

  if (!(new= (double *)Malloc (res * sizeof(double)))) {
    SeasetErrNo(SEAALLOC); return NULL; }
  
  switch (spec_type) {
  case T_CHAR:  {             /* character */
    typedef char _spec;
    _CONVERT;
    break;
  }
  case  T_SHORT:   {          /* short integer */
    typedef short _spec;
    _CONVERT;
    break;
  }
  case  T_INT:     {          /* integer */
    typedef int _spec;
    _CONVERT;
    break;
  }
  case  T_LONG:    {          /* long integer */
    typedef long _spec; 
    _CONVERT;
    break;
  }
  case  T_FLOAT:   {          /* floating point */
    typedef float _spec;
    _CONVERT;
    break;
  }
  case  T_UCHAR:   {          /* unsigned character */
    typedef unsigned char _spec;
    _CONVERT;
    break;
  }
  case  T_USHORT:  {          /* unsigned short */
    typedef unsigned short _spec;
    _CONVERT;
    break;
  }
  case  T_UINT:    {          /* unsigned integer */
    typedef unsigned int _spec;
    _CONVERT;
    break;
  }
  case  T_ULONG:   {          /* unsigned long */
    typedef unsigned long _spec;
    _CONVERT;
    break;
  }
  default:
    SeasetErrNo(SEASPCTYP);
    Free (new);
    return (NULL);
  }
  SeasetErrNo(SEANOERR);
  return(new);
}

/* ---------------------------------------------------------------------- */
char *SeaError(err_no)
     int err_no;
{
  static char *errors[] = {
  /* SEANOERR:  */ "vsSearch: no error",
  /* SEAPOINT:  */ "vsSearch: illegal NULL-pointer",
  /* SEAALLOC:  */ "vsSearch: can not allocate buffer",
  /* SEAEXIST:  */ "vsSearch: peak already exists",
  /* SEAUSER:   */ "vsSearch: error occured in user function",
  /* SEASPCTYP: */ "vsSearch: unknown spectrum type",
  /* SEANOPEAK: */ "vsSearch: no peak found",
  /* default:   */ "vsSearch: undefined error code", };

  if (err_no < sizeof(errors)) { return errors[err_no]; }
  else { return errors[sizeof(errors)-1]; }
}
/* ---------------------------------------------------------------------- */
 /*
  *  SeaPeakList peak_search(spec, err_spec, spec_type, res, left, right, 
  *  searches peaks in the whole spectrum from left to right 
  *  returns SeaPeaks in SeaPeakList or NULL on error
  */
SeaPeakList SeaPeakSearch(spec, err_spec, spec_type, res, left, right, 
			  def_prob, w_cal, cal, peak_func)
     void           *spec, *err_spec;
     int            spec_type;
     int            res;
     int            left, right;
     double         def_prob;
     double         (*w_cal)();
     Cal_t	    cal;
     int            (*peak_func)();
{
  double *new, *err_new;
  double *empty;

  if ((left == 0) && (right == 0)) {
    left = 0;
    right = res;
  }
  else if (left > right) {
    int t = left; left = right; right = t;
  }
  
  Last_Peak = (double)left;

#if BG_EXISTS
  if ((empty = empty_spectrum(res, left, right)) == NULL) {
    SeasetErrNo(SEAALLOC); return (NULL); }
#else
  empty = (double *)NULL;
#endif

  if (spec_type == T_DOUBLE) {
    SeaPeakList pl = _peak_search ((double *)err_spec, empty, (double *)spec, res, left, right, 0.0, def_prob, w_cal, cal, peak_func);
    Free (empty);
    return (pl);
  }
      
  if ((new = convert(spec, res, spec_type)) == NULL) {
    /* autofit_error is filled by convert */
    Free (empty);
    return (NULL);
  }
  if (spec == err_spec) {
    err_new = new;
  }
  else  if ((err_new = convert(err_spec, res, spec_type)) == NULL) {
    /* autofit_error is filled by convert */
    Free (new);
    Free (empty);
    return (NULL);
  }
    
  {
    SeaPeakList pl = _peak_search (err_new, empty, new, res, 
				   left, right, 0.0, def_prob,
				   w_cal, cal, peak_func);
    Free (new);
    Free (err_new);
    Free (empty);
    return (pl);
  } 
}

#undef _CONVERT

/*
 *  SeaPeakList SeaCreatePeakList()
 *  create new SeaPeakList
 *  returns NULL on Error, otherwise SeaPeakList.
 */
SeaPeakList SeaCreatePeakList()
{
  SeaPeakList peak_list;

  /*  printf("SeaCreatePeakList\n");*/
  if ((peak_list = (SeaPeakList)Malloc (sizeof (SeaPeakListRec)))) {
    peak_list->start_list   = NULL;
    peak_list->end_list     = NULL;
    peak_list->number_peaks = 0;
  }
  else {
    SeasetErrNo(SEAALLOC);
  }
  return peak_list; 
}

/*
 *  SeaPeakList SeaDupPeakList(src)
 *  duplicate SeaPeakList
 *  returns NULL on Error, otherwise SeaPeakList.
 */
SeaPeakList SeaDupPeakList(src)
     SeaPeakList src;
{
  SeaPeakList dest;

  /*  printf("SeaDupPeaklist\n");*/
  if ((dest = SeaCreatePeakList())) {
    SeaPeak run;

    ForAllSeaPeaks(src, run) {
      if (!SeaEnterPeak(dest, 
			SeaPos(run),
			SeaVol(run),
			SeaDVol(run),
			SeaProb(run),
			SeaAmpl(run),
			SeaDAmpl(run),
			SeaBg(run),
			SeaDBg(run),
			SeaChi(run),
			SeaWidth(run))) {
	SeaFreePeakList(dest); return NULL; } } }
  return dest; 
}

/*
 *  SeaPeak SeaEnterPeak (peak_list, channel, Vol, dVol, Prob,
 *  Amp, dAmp, BG, dBG, Chi, width)
 *  enter peak with given parameters to peak_list
 *  returns NULL on Error, otherwise SeaPeak.
 */
SeaPeak SeaEnterPeak (peak_list, channel, Vol, dVol, Prob,
		      Amp, dAmp, BG, dBG, Chi, width)
SeaPeakList peak_list;
double     channel;
double     Vol, dVol;
double     Prob;
double     Amp, dAmp;
double     BG, dBG;
double     Chi;
double     width;
{
  SeaPeak new;

  /*  printf("SeaEnterPeak\n");*/
  if ((new = (SeaPeak)Malloc (sizeof (SeaPeakRec))) == NULL) {
    SeasetErrNo(SEAALLOC);
    return (NULL);
  }

  if (SeaListFirst(peak_list) == NULL) {
    SeaListFirst(peak_list) = new;
    SeaListLast(peak_list)  = new;
    new->prev = NULL;
    new->next = NULL;
    Last_Peak = channel;
  }
  else if (peak_list->end_list->pos < channel) {
    new->prev = peak_list->end_list;
    new->next = NULL;
    SeaNext(SeaListLast(peak_list)) = new;
    SeaListLast(peak_list)       = new;
    Last_Peak = channel;
  }
  else {
    SeaPeak run = peak_list->start_list;
    for ( ; run->pos < channel ; run = run->next) 
      ;
    if (run->pos == channel) {
      Free(new);
      SeasetErrNo(SEAEXIST);
      return(NULL); }
    run->next->prev = new;
    run->prev->next = new;
    new->next = new;
    new->prev = run->prev;
    if (new->prev == NULL) {
      peak_list->start_list = new;
    }
  }

  ++(peak_list->number_peaks);
  SeaPos(new)	= channel;
  SeaVol(new)	= Vol;
  SeaDVol(new)	= dVol;
  SeaProb(new)	= Prob;
  SeaAmpl(new)	= Amp;
  SeaDAmpl(new)	= dAmp;
  SeaBg(new)	= BG;
  SeaDBg(new)	= dBG;
  SeaChi(new)	= Chi;
  SeaWidth(new)	= width;
  SeasetErrNo(SEANOERR);
  return (new);
}

/*
 *  SeaPeak SeaInsertPeak(peak_list, spec, err_spec, spec_type, res, channel, min_prob, w_cal, cal)
 *
 *  tries to find and fit a peak close to position channel
 *  returns SeaPeak on success, otherwise NULL
 */
SeaPeak SeaInsertPeak(peak_list, spec, err_spec, spec_type, res,
		      channel, min_prob, w_cal, cal)
     SeaPeakList    peak_list;    
     void           *spec, *err_spec;
     int            spec_type;
     int            res;
     double         channel;
     double         min_prob;
     double         (*w_cal)();
     Cal_t          cal;
{
  double *new, *err_new, *empty;
  int    left, right;

  left   = channel - 1;
  right  = channel + 1;

  if (left  < 0)    left  = 0;
  if (right >= res) right = res - 1;
  
  if ((empty = empty_spectrum(res, left, right)) == NULL) {
    SeasetErrNo(SEAALLOC); return (NULL); }
      
  if (spec_type == T_DOUBLE) {
    SeaPeak peak = _insert_peak (peak_list, (double *)err_spec, empty,
				 (double *)spec, res, left, right,
				 0.0, min_prob, w_cal, cal, NULL);
    Free (empty);
    if (peak == NULL) {
      /* autofit_error is filled by _insert_peak */
      return (NULL);
    }
    return (peak);
  }
      
  if ((new = convert((void *)spec, res,spec_type)) == NULL) {
    /* autofit_error is filled by convert */
    Free (empty);
    return (NULL);
  }

  if (spec == err_spec) {
    err_new = err_spec;
  }
  else  if ((err_new = convert((void *)err_spec, res, spec_type)) == NULL) {
    /* autofit_error is filled by convert */
    Free (new);
    Free (empty);
    return (NULL);
  }
    
  {
    SeaPeak peak = _insert_peak (peak_list, err_new, empty,
				 new, res, left, right, 0.0,
				 min_prob, w_cal, cal, NULL);
    Free (new);
    Free (err_new);
    Free (empty);
    if (peak == NULL) {
      /* autofit_error is filled by _insert_peak */
      return (NULL);
    }
    SeasetErrNo(SEANOERR);
    return (peak);
  } 
}
/* ------------------------------------------------------------------------- */
SeaPeak SeaClosestPeak(pl, chan)
SeaPeakList pl;
double      chan;
{
  SeaPeak found   = NULL;
  SeaPeak greater = NULL;
  SeaPeak run;
  
  ForAllSeaPeaks(pl, run) {
    if (SeaPos(run) > chan) {
      greater = run;
      break; } }
  if (greater == NULL) { found = SeaListLast(pl); }
  else if (greater == SeaListFirst(pl)) { found = SeaListFirst(pl); }
  else if ( (chan - SeaPos(SeaPrev(greater))) < (SeaPos(greater) - chan)) {
    found = SeaPrev(greater); }
  else {
    found = greater;
  }
  return (found);
}
/* ------------------------------------------------------------------------- */
void SeaFreePeakList(pl)
SeaPeakList pl;
{
  /*  printf("SeaFreePeakList\n");*/
  if (pl == NULL) {
    return;
  }
  {
    SeaPeak run;
    
    ForAllSeaPeaks(pl, run) {
      Free (SeaPrev(run));
    }
  }
  Free (SeaListLast(pl));
  Free(pl);
  return;
}
/* ------------------------------------------------------------------------- */
void SeaFreeClosestPeak(pl, chan)
SeaPeakList pl;
double    chan;
{
  if (pl) {
    SeaPeak peak;
    if ((peak= SeaClosestPeak(pl, chan))) {
      peak->prev->next = peak->next;
      peak->next->prev = peak->prev;
      if (peak == pl->end_list) {
	pl->end_list = peak->prev;  }
      else if (peak == pl->start_list) {
	pl->start_list = peak->next; }
      Free(peak);
      --(pl->number_peaks); } }
  return;
}
/* ------------------------------------------------------------------------- */
double SeaBgFunction(sp, x)
     SeaPeak sp;
     double x;
{
  return (double)SeaBg(sp);
}
/* ------------------------------------------------------------------------- */
double SeaFunction(sp, x)
     SeaPeak sp;
     double x;
{
  return (SeaBgFunction(sp, x) + SeaPeakFunction(sp, x));
}
/* ------------------------------------------------------------------------- */
double SeaPeakFunction(sp, x)
     SeaPeak sp;
     double x;
{
  if (SeaWidth(sp) != 0.0) {
    double e= (x - SeaPos(sp)) / SeaWidth(sp);
    return SeaAmpl(sp) * exp(-e * e * 0.5); }
  else { return 0.0; }
}
/* ------------------------------------------------------------------------- */
